#!/usr/bin/env python3

#TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty, Qt, QEvent
from PyQt5.QtGui import QMouseEvent
import PyQt5.QtWidgets

import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.toggle_ui


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class Toggle(epyqlib.widgets.abstracttxwidget.AbstractTxWidget):
    def __init__(self, parent=None, in_designer=False):
        self._frame = None
        self._signal = None
        self._value_labels_visible = True

        super().__init__(
            ui_class=epyqlib.widgets.toggle_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        self.ui.value.installEventFilter(self)
        # TODO: CAMPid 398956661298765098124690765

        self.ui.value.setFixedHeight(3 * self.ui.on.fontMetrics().height())

        self.cached_value = self.ui.value.sliderPosition()
        self.ui.value.actionTriggered.connect(self.action_triggered)
        self.ui.value.valueChanged.connect(self.slider_value_changed)

    @pyqtProperty(bool)
    def value_labels_visible(self):
        return self._value_labels_visible

    @value_labels_visible.setter
    def value_labels_visible(self, visible):
        self._value_labels_visible = bool(visible)
        self.ui.on.setVisible(self.value_labels_visible)
        self.ui.off.setVisible(self.value_labels_visible)

    def eventFilter(self, qobject, qevent):
        if isinstance(qevent, QMouseEvent) and self.tx:
            if (qevent.button() == Qt.LeftButton
                    and qevent.type() == QEvent.MouseButtonRelease
                    and self.rect().contains(qevent.localPos().toPoint())):
                self.toggle_released()

            return True

        return False

    def set_value(self, value):
        # TODO: quit hardcoding this and it's better implemented elsewhere
        if self.signal_object is not None:
            value = bool(self.signal_object.value)
        elif value is None:
            value = False
        else:
            value = bool(value)

        self.ui.value.setSliderPosition(value)

    def toggle_released(self):
        self.ui.value.triggerAction(
            PyQt5.QtWidgets.QAbstractSlider.SliderToMinimum
            if self.ui.value.sliderPosition() else
            PyQt5.QtWidgets.QAbstractSlider.SliderToMaximum
        )

    def action_triggered(self, action):
        interesting_actions = (
            PyQt5.QtWidgets.QAbstractSlider.SliderSingleStepAdd,
            PyQt5.QtWidgets.QAbstractSlider.SliderSingleStepSub,
            PyQt5.QtWidgets.QAbstractSlider.SliderToMinimum,
            PyQt5.QtWidgets.QAbstractSlider.SliderToMaximum,
        )

        new_value = self.ui.value.sliderPosition()

        if action in interesting_actions and new_value != self.cached_value:
            self.widget_value_changed(new_value)

    def slider_value_changed(self, value):
        self.cached_value = value

    def set_signal(self, signal=None, force_update=False):
        if signal is not self.signal_object or force_update:
            if signal is not None:
                self.ui.off.setText(signal.enumeration[0])
                self.ui.on.setText(signal.enumeration[1])
                signal.value_changed.connect(self.signal_value_changed)
            else:
                self.ui.off.setText('-')
                self.ui.on.setText('-')
        epyqlib.widgets.abstracttxwidget.AbstractTxWidget.set_signal(
            self, signal, force_update=force_update)

    def signal_value_changed(self, value):
        self.ui.value.setSliderPosition(bool(value))

    def set_unit_text(self, units):
        pass


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import logging
import argparse
import can
import epyqlib.busproxy
import epyqlib.canneo
import epyqlib.ticoff
import epyqlib.twisted.busproxy
import epyqlib.twisted.cancalibrationprotocol as ccp
import epyqlib.utils.twisted
import functools
import itertools
import math
import platform
import qt5reactor
import signal
import sys
import time
import twisted

from PyQt5.QtCore import QObject, pyqtSignal
from PyQt5.QtWidgets import QApplication


logger = logging.getLogger(__name__)


class Flasher(QObject):
    # TODO: CAMPid 7531968542136967546542452
    progress_messages = pyqtSignal(int)
    completed = pyqtSignal()
    done = pyqtSignal()
    failed = pyqtSignal()
    canceled = pyqtSignal()

    def __init__(self, file, bus, progress=None, retries=5, parent=None):
        super().__init__(parent)

        self.progress = progress
        self.deferred = None
        self._canceled = False

        self.completed.connect(self.done)

        self.protocol = ccp.Handler(endianness='big')
        self.protocol.messages_sent.connect(self.update_progress)
        from twisted.internet import reactor
        self.transport = epyqlib.twisted.busproxy.BusProxy(
            protocol=self.protocol,
            reactor=reactor,
            bus=bus)

        coff = epyqlib.ticoff.Coff()
        coff.from_stream(file)

        self.retries = retries

        self.sections = [s for s in coff.sections
                         if s.data is not None and s.virt_size > 0]

        self.download_bytes = sum([len(s.data) for s in self.sections])
        download_messages_to_send = sum(
            [math.ceil(len(s.data) / 6) for s in self.sections])
        # For every 5 download messages there is also 1 set MTA and 1 CRC.
        # There will likely be a couple retries and there's a bit more overhead
        # to get started.
        self.total_messages_to_send = download_messages_to_send * 7 / 5 + 15

        self.connect_to_progress()

        self._data_start_time = None
        self.data_delta_time = None

    def update_progress(self, messages_sent):
        self.progress_messages.emit(messages_sent)

    def connect_to_progress(self, progress=None):
        if progress is not None:
            self.progress = progress

        if self.progress is not None:
            self.progress.setMinimumDuration(0)
            # Default to a busy indicator, progress maximum will be set later
            self.progress.setMinimum(0)
            self.progress.setMaximum(0)
            self.progress_messages.connect(self.progress.setValue)
            self.progress.canceled.connect(self.cancel)

    def disconnect_from_progress(self):
        if self.progress is not None:
            self.progress_messages.disconnect(self.progress.setValue)
            self.progress.canceled.disconnect(self.cancel)

            self.progress = None

    def set_progress_label(self, text):
        if self.progress is not None:
            self.progress.setLabelText(text)

    def set_progress_range(self):
        if self.progress is not None:
            self.progress.setMinimum(0)
            self.progress.setMaximum(self.total_messages_to_send)

    def show_progress(self):
        if self.progress is not None:
            self.progress.show()

    def hide_progress_cancel_button(self):
        if self.progress is not None:
            self.progress.setCancelButton(None)

    def cancel(self):
        if self.deferred is not None and not self._canceled:
            self._canceled = True
            self.disconnect_from_progress()
            self.protocol.cancel()
            self.deferred.cancel()

    def flash(self):
        # We should start sending before the bootloader is listening to help
        # make sure we catch it.

        self.set_progress_label('Searching...')
        self.show_progress()

        # let any buffered/old messages get dumped
        d = epyqlib.utils.twisted.sleep(0.5)
        self.deferred = d
        d.addCallback(lambda _: epyqlib.utils.twisted.retry(
            function=lambda: self.protocol.connect(timeout=0.2),
            times=self.retries,
            acceptable=[epyqlib.utils.twisted.RequestTimeoutError])
        )

        d.addCallback(lambda _: self.set_progress_label('Clearing...'))

        # Since we will send multiple connects in most cases we should give
        # the bootloader a chance to respond to all of them before moving on.
        d.addCallback(
            lambda _: epyqlib.utils.twisted.sleep(min(1, 0.01 * self.retries))
        )
        # unlock

        d.addCallback(lambda _: epyqlib.utils.twisted.timeout_retry(
            lambda : self.protocol.set_mta(
                address_extension=ccp.AddressExtension.configuration_registers,
                address=0)
        ))
        d.addCallback(lambda _: epyqlib.utils.twisted.timeout_retry(
            lambda : self.protocol.unlock(section=ccp.Password.dsp_flash)
        ))
        d.addCallback(lambda _: epyqlib.utils.twisted.timeout_retry(
            lambda : self.protocol.set_mta(
                address_extension=ccp.AddressExtension.flash_memory,
                address=0)
        ))
        d.addCallbacks(lambda _: epyqlib.utils.twisted.timeout_retry(
            self.protocol.clear_memory,
            times=2
        ))

        d.addCallback(lambda _: self.set_progress_label('Flashing...'))
        d.addCallback(lambda _: self.set_progress_range())
        d.addCallback(lambda _: self.hide_progress_cancel_button())

        d.addCallback(lambda _: self._start_timing_data())

        self.protocol.continuous_crc = None

        for section in self.sections:
            if len(section.data) % 2 != 0:
                data = itertools.chain(section.data, [0])
            else:
                data = section.data
            callback = functools.partial(
                self.protocol.download_block,
                address_extension=ccp.AddressExtension.flash_memory,
                address=section.virt_addr,
                data=data
            )
            logger.debug('0x{:08X}'.format(section.virt_addr))
            d.addCallback(lambda _, cb=callback: cb())

        d.addCallback(lambda _: epyqlib.utils.twisted.sleep(1))
        d.addCallback(lambda _: self.protocol.build_checksum(
            checksum=self.protocol.continuous_crc, length=0))
        d.addCallback(lambda _: epyqlib.utils.twisted.sleep(1))
        d.addCallback(lambda _: self.protocol.disconnect())
        d.addCallback(lambda _: self._completed())
        d.addErrback(self._failed)

        logger.debug('---------- started')

    def _start_timing_data(self):
        self._data_start_time = time.monotonic()
        logger.debug('Started timing data at {}'.format(self._data_start_time))
        # return twisted.internet.defer.succeed()

    def _completed(self):
        self.data_delta_time = time.monotonic() - self._data_start_time
        self.completed.emit()

    def _failed(self, result):
        epyqlib.utils.twisted.logit(result)
        if self._canceled:
            self.canceled.emit()
        else:
            self.failed.emit()
        self.done.emit()


def parse_args(args):
    default = {
        'Linux': {'bustype': 'socketcan', 'channel': 'can0'},
        'Windows': {'bustype': 'pcan', 'channel': 'PCAN_USBBUS1'}
    }[platform.system()]

    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--file', '-f', type=argparse.FileType('rb'), required=True)
    parser.add_argument('--interface', '-i', default=default['bustype'])
    parser.add_argument('--channel', '-c', default=default['channel'])
    parser.add_argument('--bitrate', '-b', default=250000)

    return parser.parse_args(args)


def main(args=None):
    app = QApplication(sys.argv)

    if args is None:
        args = sys.argv[1:]

    args = parse_args(args=args)

    if args.verbose >= 1:
        logger.setLevel(logging.DEBUG)

    if args.verbose >= 2:
        twisted.internet.defer.setDebugging(True)

    if args.verbose >= 3:
        logging.getLogger().setLevel(logging.DEBUG)

    qt5reactor.install()
    from twisted.internet import reactor

    reactor.runReturn()

    QApplication.instance().aboutToQuit.connect(about_to_quit)

    real_bus = can.interface.Bus(bustype=args.interface,
                                 channel=args.channel,
                                 bitrate=args.bitrate)
    bus = epyqlib.busproxy.BusProxy(bus=real_bus, auto_disconnect=False)

    flasher = Flasher(file=args.file, bus=bus)

    flasher.completed.connect(lambda f=flasher: completed(flasher=f))
    flasher.failed.connect(failed)
    flasher.done.connect(bus.set_bus)

    flasher.flash()

    return app.exec()


def about_to_quit():
    from twisted.internet import reactor
    reactor.stop()


def completed(flasher):
    print("Flashing completed successfully")
    print('Data time: {:.3f} seconds for {} bytes or {:.0f} bytes/second'
          .format(flasher.data_delta_time,
                  flasher.download_bytes,
                  flasher.download_bytes / flasher.data_delta_time))
    QApplication.instance().quit()


def failed():
    print("Flashing failed")
    QApplication.instance().exit(1)


def _entry_point():
    import traceback

    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(message)s')

    def excepthook(excType, excValue, tracebackobj):
        logger.debug('Uncaught exception hooked:')
        traceback.print_exception(excType, excValue, tracebackobj)

    sys.excepthook = excepthook
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    return main()


if __name__ == '__main__':
    sys.exit(_entry_point())



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.check_ui


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class Check(epyqlib.widgets.abstracttxwidget.AbstractTxWidget):
    def __init__(self, parent=None, in_designer=False):
        self._frame = None
        self._signal = None

        super().__init__(
            ui_class=epyqlib.widgets.check_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        # TODO: CAMPid 398956661298765098124690765
        self.ui.value.toggled.connect(self.widget_value_changed)

    def set_value(self, value):
        # TODO: quit hardcoding this and it's better implemented elsewhere
        if self.signal_object is not None:
            value = bool(self.signal_object.value)
        elif value is None:
            value = False
        else:
            value = bool(value)

        self.ui.value.setChecked(value)

    def signal_value_changed(self, value):
        self.ui.value.setChecked(bool(value))

    def set_unit_text(self, units):
        pass


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'filesview.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(400, 300)
        self.gridLayout_2 = QtWidgets.QGridLayout(Form)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.listView = QtWidgets.QListView(Form)
        self.listView.setObjectName("listView")
        self.gridLayout.addWidget(self.listView, 1, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.something = QtWidgets.QPushButton(Form)
        self.something.setObjectName("something")
        self.horizontalLayout.addWidget(self.something)
        self.more = QtWidgets.QPushButton(Form)
        self.more.setObjectName("more")
        self.horizontalLayout.addWidget(self.more)
        self.other = QtWidgets.QPushButton(Form)
        self.other.setObjectName("other")
        self.horizontalLayout.addWidget(self.other)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        self.gridLayout_2.addLayout(self.gridLayout, 0, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.something.setText(_translate("Form", "Something"))
        self.more.setText(_translate("Form", "More"))
        self.other.setText(_translate("Form", "Other..."))



import epyqlib.filesview
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = 'Copyright 2018, EPC Power Corp.'
__license__ = 'GPLv2+'


class FilesViewPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._group = 'EPC - General'
        self._init = epyqlib.filesview.FilesViewQtBuilder
        self._module_path = 'epyqlib.filesview'
        self._name = 'FilesView'



import collections

import attr
import twisted.internet.task

import epyqlib.utils.twisted


@attr.s
class Request:
    f = attr.ib()
    period = attr.ib()


@attr.s
class Element:
    loop = attr.ib()
    period = attr.ib()


@attr.s
class Set:
    requests = attr.ib(init=False,
                       default=attr.Factory(lambda: collections.defaultdict(
                           dict)))
    periods = attr.ib(init=False, default=attr.Factory(dict))
    loops = attr.ib(init=False, default=attr.Factory(dict))

    def add_request(self, key, request):
        self.requests[request.f][key] = request

        self._update_period(request.f)

    def remove_request(self, key, request):
        self.requests[request.f].pop(key)

        self._update_period(request.f)

    def _update_period(self, f):
        minimum_period = min(r.period for r in self.requests[f].values())

        if f not in self.loops:
            self.loops[f] = Element(
                loop=twisted.internet.task.LoopingCall(f),
                period=minimum_period
            )

            _start_loop(self.loops[f].loop, self.loops[f].period)

        if minimum_period != self.loops[f].period:
            self.loops[f].loop.stop()

            self.loops[f] = Element(
                loop=twisted.internet.task.LoopingCall(f),
                period=minimum_period
            )

            _start_loop(self.loops[f].loop, self.loops[f].period)

    def stop(self):
        for element in self.loops.values():
            if element.loop.running:
                element.loop.stop()

    def start(self):
        for element in self.loops.values():
            if not element.loop.running:
                _start_loop(element.loop, element.period)


def _start_loop(loop, period):
    loop_deferred = loop.start(period)
    loop_deferred.addErrback(epyqlib.utils.twisted.errbackhook)


if __name__ == '__main__':
    # TODO: move this to a unit test and actually check it's result
    #       automatically
    import twisted.internet.reactor

    a = lambda: print('a')
    b = lambda: print('b')
    requests = (
        (1, Request(f=a, period=0.5)),
        (0, Request(f=a, period=2)),
        (0, Request(f=a, period=1)),
        (0, Request(f=b, period=1)),
    )

    looping_set = Set()
    for key, request in requests:
        print('Adding: {}: {}'.format(key, request))
        looping_set.add_request(key=key, request=request)

    reactor = twisted.internet.reactor
    reactor.callLater(5, reactor.stop)
    reactor.run()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'toggle.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(101, 42)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(Form)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.on = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.on.sizePolicy().hasHeightForWidth())
        self.on.setSizePolicy(sizePolicy)
        self.on.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.on.setObjectName("on")
        self.verticalLayout.addWidget(self.on)
        spacerItem = QtWidgets.QSpacerItem(0, 0, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.verticalLayout.addItem(spacerItem)
        self.off = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.off.sizePolicy().hasHeightForWidth())
        self.off.setSizePolicy(sizePolicy)
        self.off.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.off.setObjectName("off")
        self.verticalLayout.addWidget(self.off)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.value = QtWidgets.QSlider(Form)
        self.value.setMaximum(1)
        self.value.setOrientation(QtCore.Qt.Vertical)
        self.value.setObjectName("value")
        self.horizontalLayout.addWidget(self.value)
        self.horizontalLayout.setStretch(0, 1)
        self.horizontalLayout_2.addLayout(self.horizontalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "TextLabel"))
        self.on.setText(_translate("Form", "On"))
        self.off.setText(_translate("Form", "Off"))



import collections
import contextlib
import distutils.dir_util
import importlib
import json
import logging
import os
import shutil
import tempfile
import textwrap
import pathlib

import attr
import click
import lxml.etree

import epyqlib.deviceextension
import epyqlib.utils.general
import epyqlib.utils.click

# See file COPYING in this source tree
__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


logger = logging.getLogger(__name__)


@attr.s
class Converter:
    old_version = attr.ib()
    new_version = attr.ib()
    description = attr.ib()
    function = attr.ib()

    def __call__(self, *args, **kwargs):
        return self.function(self, *args, **kwargs)


converters = epyqlib.utils.general.Collector(Converter, 'function')


def get_ui_paths_0(device_dict):
    for ui_path_name in ['ui_path', 'ui_paths', 'menu']:
        ui_paths = device_dict.get(ui_path_name)

        if ui_paths is not None:
            break
    else:
        ui_paths = {}

    if not isinstance(ui_paths, dict):
        ui_paths = {"Dash": ui_paths}

    return ui_paths


@converters.append(
    old_version=(0,8),
    new_version=(0,9),
    description='_frame/_signal -> _signal_path_element_...')
def cf(self, source_directory, source_file_name, destination_path):
    with open(os.path.join(source_directory, source_file_name)) as f:
        device_dict = json.load(f, object_pairs_hook=collections.OrderedDict)

    ui_paths = get_ui_paths_0(device_dict)

    module_path = device_dict.get('module', None)
    if module_path is None:
        module = epyqlib.deviceextension
    else:
        spec = importlib.util.spec_from_file_location(
            'extension', os.path.join(source_directory, module_path))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

    referenced_files = (
        source_file_name,
        *(
            device_dict.get(name) for name in
            (
                'module',
                'can_path',
                'compatibility',
                'parameter_defaults',
                'parameter_hierarchy',
            )
            if name in device_dict
        ),
        *ui_paths.values(),
        *module.referenced_files(device_dict),
    )

    referenced_files = {f: False for f in referenced_files}

    for ui in referenced_files:
        if os.path.splitext(ui)[1].casefold() != '.ui':
            continue

        with open(os.path.join(source_directory, ui)) as f:
            root = lxml.etree.parse(f)

        widgets = root.xpath(
            "//widget[property[re:match(@name, '(_|^)(signal|frame)')]]",
            namespaces={"re": "http://exslt.org/regular-expressions"},
        )

        for widget in widgets:
            logging.debug('Widget: {}'.format(widget.get('name')))
            signal_paths = widget.xpath(
                "property[re:match(@name, '(_|^)(signal|frame)')]",
                namespaces={"re": "http://exslt.org/regular-expressions"},
            )
            full_names = tuple(e.get('name') for e in signal_paths)

            signal_path_names = set()
            for name in full_names:
                left, underscore, right = name.rpartition('_')
                if len(underscore) > 0:
                    signal_path_names.add(left)
                else:
                    signal_path_names.add(right)

            for name in signal_path_names:
                logging.debug('  Signal path name: {}'.format(name))
                xpath = "property[contains(@name, '{}')]".format(name)
                properties = widget.xpath(xpath)
                logging.debug('    Properties: {}'.format(
                    tuple(e.get('name') for e in properties),
                ))

                for i, p in enumerate(properties):
                    left, underscore, right = p.get('name').rpartition('_')
                    p.set(
                        'name',
                        ''.join((
                            left,
                            underscore,
                            'signal_path_element_',
                            str(i),
                        ))
                    )

        with open(os.path.join(destination_path, ui), 'wb') as f:
            root.write(f, xml_declaration=True, encoding='UTF-8')
            f.write('\n'.encode('utf-8'))

        logging.info('Updated {}'.format(ui))
        referenced_files[ui] = True

    device_dict['format_version'] = self.new_version
    device_dict.move_to_end('format_version', last=False)

    destination_device_file = os.path.join(destination_path, source_file_name)
    with open(destination_device_file, 'w') as f:
        json.dump(device_dict, f, indent=4)
        f.write('\n')
    referenced_files[source_file_name] = True

    for f, handled in referenced_files.items():
        if not handled:
            logging.info('Copying {}'.format(f))
            shutil.copy(
                src=os.path.join(source_directory, f),
                dst=destination_path,
            )

    return destination_device_file


@converters.append(
    old_version=(0,9),
    new_version=(1,),
    description='signal_path_element_0,1,2 to ; delimited signal_path '
                'properties in .ui')
def cf(self, source_directory, source_file_name, destination_path):
    with open(os.path.join(source_directory, source_file_name)) as f:
        device_dict = json.load(f, object_pairs_hook=collections.OrderedDict)

    ui_paths = get_ui_paths_0(device_dict)

    referenced_files = (
        source_file_name,
        *(
            device_dict.get(name) for name in
            (
                'module',
                'can_path',
                'compatibility',
                'parameter_defaults',
                'parameter_hierarchy',
            )
            if name in device_dict
        ),
        *ui_paths.values(),
    )

    referenced_files = {f: False for f in referenced_files}

    module = device_dict.get('module')
    if module is not None:
        referenced_files_string = textwrap.dedent('''\
        def referenced_files(raw_dict):
            return ()
        ''')

        with open(os.path.join(source_directory, module)) as f_in:
            with open(os.path.join(destination_path, module), 'w') as f_out:
                written = any(
                    l.startswith('def referenced_files(') for l in f_in
                )

                f_in.seek(0)

                for line in f_in:
                    if not written and line.startswith(('class ', 'def ')):
                        logging.info('Adding referenced_files() to module')

                        f_out.write(referenced_files_string)
                        f_out.write('\n\n')

                        written = True

                    f_out.write(line)

                if not written:
                    f_out.write('\n\n')
                    f_out.write(referenced_files_string)

        referenced_files[module] = True

    ui_files = tuple(
        f
        for f in referenced_files
        if f.endswith('.ui')
    )

    for ui in ui_files:
        with open(os.path.join(source_directory, ui)) as f:
            root = lxml.etree.parse(f)

        widgets = root.xpath(
            "//widget[property[contains(@name, '_path_element_0')]]"
        )

        for widget in widgets:
            logging.debug('Widget: {}'.format(widget.get('name')))
            signal_paths = widget.xpath(
                "property[contains(@name, '_path_element_')]")
            full_names = tuple(e.get('name') for e in signal_paths)
            signal_path_names = set(
                name.rpartition('_element_')[0] for name in full_names
            )
            for name in signal_path_names:
                logging.debug('  Signal path name: {}'.format(name))
                xpath = "property[contains(@name, '{}')]".format(name)
                properties = widget.xpath(xpath)
                logging.debug('    Properties: {}'.format(
                    tuple(e.get('name') for e in properties),
                ))

                modify, to_remove = properties[0], properties[1:]
                strings = widget.xpath(xpath + '/string')
                modify.set('name', modify.get('name').rpartition('_element')[0])
                modify[0].text = ';'.join(s.text for s in strings)
                for e in to_remove:
                    e.getparent().remove(e)

        with open(os.path.join(destination_path, ui), 'wb') as f:
            root.write(f, xml_declaration=True, encoding='UTF-8')
            f.write('\n'.encode('utf-8'))

        logging.info('Updated {}'.format(ui))
        referenced_files[ui] = True

    device_dict['format_version'] = self.new_version
    device_dict.move_to_end('format_version', last=False)

    destination_device_file = os.path.join(destination_path, source_file_name)
    with open(destination_device_file, 'w') as f:
        json.dump(device_dict, f, indent=4)
        f.write('\n')
    referenced_files[source_file_name] = True

    for f, handled in referenced_files.items():
        if not handled:
            logging.info('Copying {}'.format(f))
            shutil.copy(
                src=os.path.join(source_directory, f),
                dst=destination_path,
            )

    return destination_device_file


@converters.append(
    old_version=(1,),
    new_version=(2,),
    description='self.ui is not longer self',
)
def cf(self, source_directory, source_file_name, destination_path):
    with open(os.path.join(source_directory, source_file_name)) as f:
        device_dict = json.load(f, object_pairs_hook=collections.OrderedDict)

    ui_paths = get_ui_paths_0(device_dict)

    module_path = device_dict.get('module', None)

    if module_path is None:
        module = epyqlib.deviceextension
    else:
        spec = importlib.util.spec_from_file_location(
            'extension', os.path.join(source_directory, module_path))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

    referenced_files = (
        source_file_name,
        *(
            device_dict.get(name) for name in
            (
                'module',
                'can_path',
                'compatibility',
                'parameter_defaults',
                'parameter_hierarchy',
            )
            if name in device_dict
        ),
        *ui_paths.values(),
        *module.referenced_files(device_dict),
    )

    referenced_files = {f: False for f in referenced_files}

    with open(os.path.join(source_directory, module_path), 'r') as f:
        source = f.read()

    replacements = [
        [
            'd.voltage_droop_mode.ui.signal_object',
            'd.voltage_droop_mode.signal_object',
        ],
        [
            'd.voltage_droop_mode.ui.setVisible',
            'd.voltage_droop_mode.setVisible',
        ],
        [
            'self.setup.pump_reset_button.value',
            'self.setup.pump_reset_button.ui.value',
        ],
    ]

    for original, replacement in replacements:
        source = source.replace(original, replacement)

    with open(os.path.join(destination_path, module_path), 'w') as f:
        f.write(source)

    referenced_files[module_path] = True

    device_dict['format_version'] = self.new_version
    device_dict.move_to_end('format_version', last=False)

    destination_device_file = os.path.join(destination_path, source_file_name)
    with open(destination_device_file, 'w') as f:
        json.dump(device_dict, f, indent=4)
        f.write('\n')
    referenced_files[source_file_name] = True

    for f, handled in referenced_files.items():
        if not handled:
            logging.info('Copying {}'.format(f))
            shutil.copy(
                src=os.path.join(source_directory, f),
                dst=destination_path,
            )

    return destination_device_file


@converters.append(
    old_version=(2,),
    new_version=None,
    description='template for next format version')
def cf(self, source_directory, source_file_name, destination_path):
    with open(os.path.join(source_directory, source_file_name)) as f:
        device_dict = json.load(f, object_pairs_hook=collections.OrderedDict)

    ui_paths = get_ui_paths_0(device_dict)

    module_path = device_dict.get('module', None)
    if module_path is None:
        module = epyqlib.deviceextension
    else:
        spec = importlib.util.spec_from_file_location(
            'extension', os.path.join(source_directory, module_path))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

    referenced_files = (
        source_file_name,
        *(
            device_dict.get(name) for name in
            (
                'module',
                'can_path',
                'compatibility',
                'parameter_defaults',
                'parameter_hierarchy',
            )
            if name in device_dict
        ),
        *ui_paths.values(),
        *module.referenced_files(device_dict),
    )

    referenced_files = {f: False for f in referenced_files}

    device_dict['format_version'] = self.new_version
    device_dict.move_to_end('format_version', last=False)

    destination_device_file = os.path.join(destination_path, source_file_name)
    with open(destination_device_file, 'w') as f:
        json.dump(device_dict, f, indent=4)
        f.write('\n')
    referenced_files[source_file_name] = True

    for f, handled in referenced_files.items():
        if not handled:
            logging.info('Copying {}'.format(f))
            shutil.copy(
                src=os.path.join(source_directory, f),
                dst=destination_path,
            )

    return destination_device_file


def just_copy(source_directory, source_file_name, destination_path):
    distutils.dir_util.copy_tree(source_directory, destination_path)


def conversion_summaries():
    return tuple(
        '{} -> {}: {}'.format(c.old_version, c.new_version, c.description)
        for c in sorted(
            converters, key=lambda c: (c.old_version, c.new_version)
        )
    )


def version(path):
    with open(path) as f:
        device_dict = json.load(f, object_pairs_hook=collections.OrderedDict)

    v = device_dict.get('format_version', None)

    ui_paths = get_ui_paths_0(device_dict)

    module_path = device_dict.get('module', None)
    if module_path is None:
        module = epyqlib.deviceextension
    else:
        spec = importlib.util.spec_from_file_location(
            'extension', os.path.join(os.path.dirname(path), module_path))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

    referenced_files = (
        path,
        *(
            device_dict.get(name) for name in
            (
                'module',
                'can_path',
                'compatibility',
                'parameter_defaults',
                'parameter_hierarchy',
            )
            if name in device_dict
        ),
        *ui_paths.values(),
    )

    try:
        referenced_files += tuple(module.referenced_files(device_dict))
    except:
        pass

    # path might be a string or a pathlib.Path object
    # need to be able to handle both 
    referenced_files = tuple(pathlib.Path(path) for path in referenced_files)

    ui_files = tuple(
        f
        for f in referenced_files
        if f.suffix == '.ui'
    )

    if v is None:
        fragment = '_path_element_'

        for ui in ui_files:
            with open(os.path.join(os.path.dirname(path), ui)) as f:
                if any(fragment in line for line in f):
                    v = (0,9)
                    break

    if v is None:
        v = (0,8)

    return tuple(v)


# TODO: calculate by largest version number in converts?
current_version = max(
    c.new_version for c in converters if c.new_version is not None
)


def is_latest(path):
    return version(path) == current_version


def get_converter(old_version, new_version=current_version):
    converter, = (
        c for c in converters
        if c.old_version == old_version and c.new_version == new_version
    )

    return converter


def get_converters(old_version=None, new_version=None):
    return tuple(
        c for c in converters
        if (
            (c.old_version == old_version or old_version is None)
            and (c.new_version == new_version or new_version is None)
        )
    )


def get_converter_chain(old_version, new_version=current_version):
    converter_chain = []

    old = old_version

    while True:
        cs = get_converters(old_version=old)
        converter = max(cs, key=lambda c: c.new_version)
        converter_chain.append(converter)
        old = converter.new_version

        if converter.new_version == new_version:
            break

    return converter_chain


def convert(source_path, destination_path, destination_version=None):
    if destination_version is None:
        destination_version = current_version

    source_version = version(source_path)

    chain = get_converter_chain(source_version, destination_version)
    chain.append(just_copy)

    temporary_directories = tuple(
        tempfile.TemporaryDirectory() for _ in range(len(chain) - 1)
    )

    directories = (
        os.path.dirname(source_path),
        *(d.name for d in temporary_directories),
        destination_path,
    )

    directory_pairs = tuple(epyqlib.utils.general.pairwise(directories))

    for c, (source, destination) in zip(chain, directory_pairs):
        if c is not just_copy:
            logging.info(
                'Converting {} -> {}'.format(c.old_version, c.new_version)
            )

        c(
            source_directory=source,
            source_file_name=os.path.basename(source_path),
            destination_path=destination,
        )

    for temporary_directory in temporary_directories:
        temporary_directory.cleanup()

    logging.info('New device file saved to: {}'.format(destination_path))

    return os.path.join(destination_path, os.path.basename(source_path))


@contextlib.contextmanager
def updated(path):
    if epyqlib.updateepc.is_latest(path):
        yield pathlib.Path(path)
        return

    with tempfile.TemporaryDirectory() as converted_directory:
        final_file = convert(
            source_path=path,
            destination_path=converted_directory,
        )

        yield pathlib.Path(final_file)


@click.command()
@epyqlib.utils.click.verbose_option
@click.option('--target-version', '-t', default=None)
@click.argument('epc', type=click.Path(exists=True), required=True)
@click.argument('destination', type=click.Path(exists=True),
                required=True)
def main(epc, destination, target_version):
    if target_version is not None:
        target_version = tuple(
            int(s) for s in target_version.rstrip('0.').split('.')
        )

    epc_version = version(epc)
    if epc_version == current_version:
        logging.warning('Already at latest version (v{})'.format(
                '.'.join(str(s) for s in epc_version)
        ))
    else:
        convert(
            source_path=epc,
            destination_path=destination,
            destination_version=target_version,
        )



import zipfile

import click

import epyqlib.autodevice.build


@click.group()
def cli():
    pass


@cli.group()
def create():
    pass


@create.command()
@click.option(
    '--zip',
    type=click.Path(dir_okay=False, resolve_path=True),
    required=True,
)
def template(zip):
    epyqlib.autodevice.build.create_template_archive(zip)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'scriptingview.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(743, 550)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.stop_button = QtWidgets.QPushButton(Form)
        self.stop_button.setObjectName("stop_button")
        self.gridLayout.addWidget(self.stop_button, 0, 9, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 7, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 0, 3, 1, 1)
        self.run_button = QtWidgets.QPushButton(Form)
        self.run_button.setObjectName("run_button")
        self.gridLayout.addWidget(self.run_button, 0, 4, 1, 1)
        self.load_button = QtWidgets.QPushButton(Form)
        self.load_button.setObjectName("load_button")
        self.gridLayout.addWidget(self.load_button, 0, 1, 1, 1)
        self.pause_button = QtWidgets.QPushButton(Form)
        self.pause_button.setObjectName("pause_button")
        self.gridLayout.addWidget(self.pause_button, 0, 8, 1, 1)
        self.loop_button = QtWidgets.QPushButton(Form)
        self.loop_button.setObjectName("loop_button")
        self.gridLayout.addWidget(self.loop_button, 0, 5, 1, 1)
        self.continue_button = QtWidgets.QPushButton(Form)
        self.continue_button.setObjectName("continue_button")
        self.gridLayout.addWidget(self.continue_button, 0, 6, 1, 1)
        self.save_button = QtWidgets.QPushButton(Form)
        self.save_button.setObjectName("save_button")
        self.gridLayout.addWidget(self.save_button, 0, 2, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem2, 0, 0, 1, 1)
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem3, 0, 10, 1, 1)
        self.csv_edit = QtWidgets.QPlainTextEdit(Form)
        self.csv_edit.setObjectName("csv_edit")
        self.gridLayout.addWidget(self.csv_edit, 1, 0, 1, 11)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.stop_button.setText(_translate("Form", "Stop"))
        self.run_button.setText(_translate("Form", "Run"))
        self.load_button.setText(_translate("Form", "Load..."))
        self.pause_button.setText(_translate("Form", "Pause"))
        self.loop_button.setText(_translate("Form", "Loop"))
        self.continue_button.setText(_translate("Form", "Continue"))
        self.save_button.setText(_translate("Form", "Save..."))



#!/usr/bin/env python3

import argparse
import collections
import epyqlib.device
import epyqlib.utils.general
try:
    import git
except ImportError as e:
    raise ImportError('Package gitpython expected but not found') from e
import json
import os
import pip
import shutil
import stat
import sys
import tempfile
import traceback
import zipfile


# TODO: CAMPid 0238493420143087667542054268097120437916848
# http://stackoverflow.com/a/21263493/228539
def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    if os.path.isdir(name):
        os.rmdir(name)
    else:
        os.remove(name)


def collect_blobs(tree):
    blobs = tree.blobs
    for subtree in tree.trees:
        blobs.extend(collect_blobs(subtree))

    return blobs


class DeviceLoadError(Exception):
    pass


def collect(devices, output_directory, dry_run, groups=None, device_path=None,
            in_repo=True):
    if groups is None:
        groups = []

    with tempfile.TemporaryDirectory() as checkout_dir:
        os.makedirs(output_directory, exist_ok=True)

        all_urls = set()
        all_repos = []
        all_remotes = set()
        all_devices = set()

        dirs = []

        for name, values in devices.items():
            repository = values.get('repository', None)
            print('  Handling {}'.format(name))
            if repository is not None:
                dir = os.path.join(checkout_dir, name)
                dirs.append(dir)

                print('    Cloning {}'.format(values['repository']))
                repo = git.Repo.clone_from(values['repository'], dir)
                all_repos.append(repo)
                repo.git.checkout(values['branch'])
                sha = repo.head.commit.hexsha

                if values['repository'] not in all_urls:
                    all_urls.add(values['repository'])
                    all_remotes.add(repo)
                all_devices.add((values['repository'], values['branch'], values['file']))
            else:
                dir = os.path.dirname(device_path)

                if in_repo:
                    repo = git.Repo(dir, search_parent_directories=True)

                    sha = '{dirty}-{sha}'.format(
                        sha=repo.head.commit.hexsha,
                        dirty='local{}'.format('_dirty' if repo.is_dirty() else ''),
                    )
                else:
                    sha = 'not_applicable'

            device_groups = values.get('groups', [])
            if set(device_groups).isdisjoint(set(groups)):
                print('    Skipping {}'.format(values['file']))
                print('        Device groups {} not in selected groups {}'.format(device_groups, groups))
                continue

            device_path = os.path.join(dir, values['file'])
            print('    Loading {}'.format(values['file']))
            try:
                device = epyqlib.device.Device(file=device_path,
                                               only_for_files=True)
            except Exception as e:
                try:
                    raise DeviceLoadError("Unable to open '{}'".format(device_path)) from e
                except:
                    print()
                    traceback.print_exc()
                    print()
                    continue
            device_dir = os.path.dirname(device_path)
            device_file_name = os.path.basename(device_path)
            referenced_files = [
                device_file_name,
                *device.referenced_files
            ]

            zip_file_name = name + '.epz'
            zip_path = os.path.join(output_directory, zip_file_name)
            print('    {} {}'.format({False: 'Writing', True: 'Not writing'}[dry_run],
                                     zip_file_name))

            if not dry_run:
                epyqlib.utils.general.write_device_to_zip(
                    zip_path=zip_path,
                    checkout_dir=checkout_dir,
                    epc_dir=device_dir,
                    referenced_files=referenced_files,
                    sha=sha
                )

        all_devices_strings = ['{}:{}:{}'.format(url, branch, file) for url, branch, file in all_devices]

        other_devices = set()

        for repo in all_remotes:
            origin_heads = [
                r for r in repo.refs
                if str(r).startswith('origin/')
            ]

            for branch in origin_heads:
                tree = repo.tree(str(branch))
                blobs = collect_blobs(tree=tree)
                for blob in blobs:
                    if blob.path.endswith('.epc'):
                        s = '{}:{}:{}'.format(repo.remotes.origin.url, str(branch)[len('origin/'):], blob.path)
                        if s not in all_devices_strings:
                            other_devices.add(s)

        other_devices = sorted(other_devices)
        if len(other_devices) > 0:
            print()
            print('Other devices available from the referenced repositories:')
            for device in other_devices:
                print('    {}'.format(device))

        # TODO: workaround for 'bug' in gitpython
        #       https://epc-phab.exana.io/T407
        #       https://github.com/gitpython-developers/GitPython/issues/546
        for repo in all_repos:
            repo.git.clear_cache()

        for dir in dirs:
            # http://bugs.python.org/issue26660
            shutil.rmtree(dir, onerror=del_rw)


class LoadJsonFiles(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, [])

        for value in values:
            getattr(namespace, self.dest).append(
                (value.name, json.loads(value.read(), object_pairs_hook=collections.OrderedDict))
            )

        setattr(namespace, self.dest + '_path', os.path.abspath(value.name))


def parse_args(args):
    parser =  argparse.ArgumentParser()
    parser.add_argument('--device-file', '-d', type=argparse.FileType('r'),
                        action=LoadJsonFiles, dest='device_files', nargs='+',
                        required=True)
    parser.add_argument('--output-directory', '-o', default=os.getcwd())
    parser.add_argument('--dry-run', '-n', action='store_true')
    parser.add_argument('--group', '-g', dest='groups', action='append', required=True)

    return parser.parse_args(args)


def main(args=None, device_files=None, output_directory=None, groups=None):
    if args is None:
        args = sys.argv[1:]

    if device_files is not None:
        for device in device_files:
            args.extend(['-d', device])

    if output_directory is not None:
        args.extend(['-o', output_directory])

    if groups is not None:
        for group in groups:
            args.extend(['-g', group])

    args = parse_args(args=args)

    for name, devices in args.device_files:
        print('Processing {}'.format(name))
        collect(devices=devices,
                output_directory=args.output_directory,
                dry_run=args.dry_run,
                groups=args.groups,
                device_path=args.device_files_path)


if __name__ == '__main__':
    os.environ['PATH'] = os.pathsep.join((
        os.environ['PATH'],
        os.path.join('c:/', 'Program Files', 'Git', 'bin')
    ))

    sys.exit(main())



# This file has been generated


build_system = 'AppVeyor'
build_id = '29930071'
build_number = '435'
build_version = '1.0.435'
job_id = 'ujeqneva89n208bx'
job_url = 'https://ci.appveyor.com/project/KyleAltendorf/stlib/build/job/ujeqneva89n208bx'



import pathlib

import click
import PyQt5.uic


@click.command()
@click.option(
    '--ui',
    'ui_paths',
    multiple=True, type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    '--directory',
    '--dir',
    'directories',
    default=['.'],
    multiple=True,
    type=click.Path(exists=True, file_okay=False),
)
@click.option('--suffix', default='_ui')
@click.option('--encoding', default='utf-8')
def cli(ui_paths, directories, suffix, encoding):
    ui_paths = [pathlib.Path(path) for path in ui_paths]

    for directory in directories:
        path = pathlib.Path(directory)
        found_paths = path.rglob('*.ui')
        ui_paths.extend(found_paths)

    for path in ui_paths:
        in_path = path
        out_path = path.with_name(f'{path.stem}{suffix}.py')

        click.echo(f'Converting: {in_path} -> {out_path}')
        with open(out_path, 'w', encoding=encoding) as out_file:
            PyQt5.uic.compileUi(in_path, out_file)



import arrow
import attr
import graham
import marshmallow

import epyqlib.attrsmodel
import epyqlib.treenode


def create_blank(model=None):
    fault_log = FaultLog(model=model)
    _post_load(fault_log)

    return fault_log


def _post_load(fault_log, root=None):
    if root is None:
        root = Root()

    if fault_log.model is None:
        fault_log.model = epyqlib.attrsmodel.Model(
            root=root,
            columns=columns,
        )


@graham.schemify(tag='event')
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Event(epyqlib.treenode.TreeNode):
    time = attr.ib(
        converter=arrow.get,
        default=attr.Factory(arrow.utcnow),
    )
    graham.attrib(
        attribute=time,
        field=marshmallow.fields.DateTime(),
    )
    epyqlib.attrsmodel.attrib(
        attribute=time,
        editable=False,
    )

    value = attr.ib(
        converter=epyqlib.attrsmodel.to_int_or_none,
        default=None,
    )
    epyqlib.attrsmodel.attrib(
        attribute=value,
        editable=False,
    )

    description = attr.ib(
        converter=str,
        default=None,
    )
    epyqlib.attrsmodel.attrib(
        attribute=description,
        editable=False,
    )

    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    @staticmethod
    def can_delete(node=None):
        return False

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


Root = epyqlib.attrsmodel.Root(
    default_name='Event Log',
    valid_types=(Event,),
)

types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Event,
    ),
)


@attr.s
class FaultLog:
    model = attr.ib(default=None)

    def connect(
            self,
            process_frames,
            process_message_names,
            nv_frames,
            nv_message_names,
    ):
        process_signals = []
        for name in process_message_names:
                frame = process_frames.frame_by_name(name)
                process_signals.extend(frame.signals)

        for signal in process_signals:
            description = ':'.join((signal.frame.name, signal.name))

            def changed(value, *, signal=signal, description=description):
                self.changed(
                    value=value,
                    signal=signal,
                    description=description,
                )

            signal.value_changed.connect(changed)

        nv_signals = [
            nv
            for nv in nv_frames.all_nv()
            if nv.frame.mux_name in nv_message_names
        ]

        for signal in nv_signals:
            description = ':'.join((signal.frame.mux_name, signal.name))

            def changed(value, *, signal=signal, description=description):
                self.changed(
                    value=value,
                    signal=signal,
                    description=description,
                )

            signal.value_changed.connect(changed)

    def changed(self, signal, value, description):
        event = Event(
            value=value,
            description=description,
        )
        self.model.root.append_child(event)


# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    merge('time', Event),
    merge('value', Event),
    merge('description', Event),
    merge('uuid', *types.types.values()),
)



#!/usr/bin/env python3

#TODO: """DocString if there is one"""

import epyqlib.utils.qt
import epyqlib.widgets.compactepc_ui
import epyqlib.widgets.epc


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class CompactEpc(epyqlib.widgets.epc.Epc):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(
            parent=parent,
            ui_class=epyqlib.widgets.compactepc_ui.Ui_Form,
            in_designer=in_designer,
        )


if __name__ == '__main__':
    import sys

    print('No script functionality here')
    sys.exit(1)     # non-zero is a failure



import enum
import functools
import io
import os
import signal
import sys
import textwrap
import time
import traceback
import uuid
import weakref

import epyqlib.utils.general

import attr
from PyQt5 import QtCore
from PyQt5 import QtWidgets
import PyQt5.uic
import twisted.internet.defer

__copyright__ = 'Copyright 2017, EPC Power Corp.'
__license__ = 'GPLv2+'


# TODO: CAMPid 953295425421677545429542967596754
log = os.path.join(os.getcwd(), 'epyq.log')


_version_tag = None
_build_tag = None
_parent = None


def exception_message_box_register_versions(version_tag, build_tag):
    global _version_tag
    global _build_tag

    _version_tag = version_tag
    _build_tag = build_tag


def exception_message_box_register_parent(parent):
    global _parent

    _parent = parent


def exception_message_box(excType=None, excValue=None, tracebackobj=None):
    epyqlib.utils.general.exception_logger(excType, excValue, tracebackobj)

    def join(iterable):
        return ''.join(iterable).strip()

    expected = isinstance(excValue, epyqlib.utils.general.ExpectedException)

    if expected:
        brief = excValue.expected_message()
    else:
        brief = join(traceback.format_exception_only(
            etype=excType,
            value=excValue
        ))

    extended = join(traceback.format_exception(
        etype=excType,
        value=excValue,
        tb=tracebackobj,
    ))

    if expected:
        box = raw_exception_message_box
    else:
        box = custom_exception_message_box

    box(
        brief=brief,
        extended=extended,
        stderr=False,
    )


def custom_exception_message_box(brief, extended='', **kwargs):
    email = "kyle.altendorf@epcpower.com"

    brief = textwrap.dedent('''\
        An unhandled exception occurred. Please report the problem via email to:
                        {email}

        {brief}''').format(
        email=email,
        brief=brief,
    )

    raw_exception_message_box(brief=brief, extended=extended, **kwargs)


def raw_exception_message_box(brief, extended, stderr=True):
    version = ''
    if _version_tag is not None:
        version = 'Version Tag: {}'.format(_version_tag)

    build = ''
    if _build_tag is not None:
        build = 'Build Tag: {}'.format(_build_tag)

    info = (version, build)
    info = '\n'.join(s for s in info if len(s) > 0)
    if len(info) > 0:
        info += '\n\n'

    time_string = time.strftime("%Y-%m-%d, %H:%M:%S %Z")

    details = textwrap.dedent('''\
        {info}A log has been written to "{log}".
        {time_string}''').format(
        info=info,
        log=log,
        time_string=time_string,
    )

    if len(extended) > 0:
        details = '\n'.join(s.strip() for s in (details, '-' * 70, extended))

    if stderr:
        sys.stderr.write('\n'.join((brief, details, '')))

    dialog(
        parent=_parent,
        title='Exception',
        message=brief,
        details=details,
        icon=QtWidgets.QMessageBox.Critical,
    )


# http://stackoverflow.com/a/35902894/228539
def message_handler(mode, context, message):
    mode_strings = {
        QtCore.QtInfoMsg: 'INFO',
        QtCore.QtWarningMsg: 'WARNING',
        QtCore.QtCriticalMsg: 'CRITICAL',
        QtCore.QtFatalMsg: 'FATAL'
    }

    mode = mode_strings.get(mode, 'DEBUG')

    print('qt_message_handler: f:{file} l:{line} f():{function}'.format(
        file=context.file,
        line=context.line,
        function=context.function
    ))
    print('  {}: {}\n'.format(mode, message))


class Progress(QtCore.QObject):
    # TODO: CAMPid 7531968542136967546542452
    updated = QtCore.pyqtSignal(int)
    completed = QtCore.pyqtSignal()
    done = QtCore.pyqtSignal()
    failed = QtCore.pyqtSignal()
    canceled = QtCore.pyqtSignal()

    default_progress_label = (
        '{elapsed} seconds elapsed, {remaining} seconds remaining'
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.completed.connect(self.done)
        self.failed.connect(self.done)
        self.canceled.connect(self.done)

        self.done.connect(self._done)

        self.progress = None
        self.average = None
        self.average_timer = QtCore.QTimer()
        self.average_timer.setInterval(200)
        self.average_timer.timeout.connect(self._update_time_estimate)
        self._label_text_replace = None
        self._start_time = None

    def _done(self):
        self.average_timer.stop()
        self.average = None

        self.updated.disconnect(self.progress.setValue)
        self.progress.close()
        self.progress.deleteLater()
        self.progress = None

        self._start_time = None

    def _update_time_estimate(self):
        remaining = self.average.remaining_time(self.progress.maximum())
        try:
            remaining = round(remaining)
        except:
            pass
        self.progress.setLabelText(self._label_text_replace.format(
                elapsed=round(time.monotonic() - self._start_time),
                remaining=remaining
            )
        )

    def connect(self, progress, label_text=None):
        self.progress = progress

        if label_text is None:
            label_text = self.default_progress_label
        self._label_text_replace = label_text

        self.progress.setMinimumDuration(0)
        self.progress.setValue(0)
        # Default to a busy indicator, progress maximum can be set later
        self.progress.setMinimum(0)
        self.progress.setMaximum(0)
        self.updated.connect(self.progress.setValue)

        if self._start_time is None:
            self._start_time = time.monotonic()

        self.average = epyqlib.utils.general.AverageValueRate(seconds=30)
        self.average_timer.start()

    def configure(self, minimum=0, maximum=0):
        self.progress.setMinimum(minimum)
        self.progress.setMaximum(maximum)

    def complete(self, message=None):
        if message is not None:
            QtWidgets.QMessageBox.information(self.progress, 'EPyQ', message)

        self.completed.emit()

    def elapsed(self):
        return time.monotonic() - self._start_time

    def fail(self):
        self.failed.emit()

    def update(self, value):
        self.average.add(value)
        self.updated.emit(value)


def complete_filter_type(extension):
    if extension == '*':
        return extension

    return '*.' + extension

def create_filter_string(name, extensions):
    return '{} ({})'.format(
        name,
        ' '.join((complete_filter_type(e) for e in extensions)),
     )


def file_dialog(
        filters,
        default=0,
        save=False,
        multiple=False,
        caption='',
        parent=None,
        path_factory=str,
        **kwargs,
):
    # TODO: CAMPid 9857216134675885472598426718023132
    # filters = [
    #     ('EPC Packages', ['epc', 'epz']),
    #     ('All Files', ['*'])
    # ]
    # TODO: CAMPid 97456612391231265743713479129

    if save:
        multiple = False

    filter_strings = [create_filter_string(f[0], f[1]) for f in filters]
    filter_string = ';;'.join(filter_strings)

    if save:
        dialog = QtWidgets.QFileDialog.getSaveFileName
    elif multiple:
        dialog = QtWidgets.QFileDialog.getOpenFileNames
    else:
        dialog = QtWidgets.QFileDialog.getOpenFileName

    if 'dir' in kwargs:
        kwargs['directory'] = kwargs.pop('dir')

    selected = dialog(
        parent=parent,
        filter=filter_string,
        initialFilter=filter_strings[default],
        caption=caption,
        **kwargs
    )[0]

    if multiple:
        return [path_factory(path) for path in selected]

    if len(selected) == 0:
        return None

    return path_factory(selected)


def get_code():
    code = None

    code_file = QtCore.QFile(':/code')
    if code_file.open(QtCore.QIODevice.ReadOnly):
        code = bytes(code_file.readAll())
        code = code.decode('ascii').strip().encode('ascii')
        code_file.close()

    return code


def progress_dialog(parent=None, cancellable=False):
    progress = QtWidgets.QProgressDialog(parent)
    flags = progress.windowFlags()
    flags &= ~QtCore.Qt.WindowContextHelpButtonHint
    flags &= ~QtCore.Qt.WindowCloseButtonHint
    progress.setWindowFlags(flags)
    progress.setWindowModality(QtCore.Qt.WindowModal)
    progress.setAutoReset(False)
    if not cancellable:
        progress.setCancelButton(None)
    progress.setMinimumDuration(0)
    progress.setMinimum(0)
    progress.setMaximum(0)

    return progress


class FittedTextBrowser(QtWidgets.QTextBrowser):
    def sizeHint(self):
        default = super().sizeHint()

        if not default.isValid():
            return default

        document_size = self.document().size()

        desktops = QtWidgets.QApplication.desktop()
        screen_number = desktops.screenNumber(self.parent())
        geometry = desktops.screenGeometry(screen_number)

        if document_size.width() == 0:
            document_size.setWidth(geometry.width() * 0.25)
        if document_size.height() == 0:
            document_size.setHeight(geometry.height() * 0.4)

        scrollbar_width = QtWidgets.QApplication.style().pixelMetric(
            QtWidgets.QStyle.PM_ScrollBarExtent
        )

        width = sum((
            document_size.width(),
            self.contentsMargins().left(),
            self.contentsMargins().right(),
            scrollbar_width,
        ))

        height = sum((
            document_size.height(),
            self.contentsMargins().top(),
            self.contentsMargins().bottom(),
            scrollbar_width,
        ))

        return QtCore.QSize(width, height)


class DialogUi:
    def __init__(self, parent):
        self.layout = QtWidgets.QGridLayout(parent)
        self.icon = QtWidgets.QLabel(parent)
        self.message = FittedTextBrowser(parent)
        self.details = FittedTextBrowser(parent)
        self.copy = QtWidgets.QPushButton(parent)
        self.save = QtWidgets.QPushButton(parent)
        self.show_details = QtWidgets.QPushButton(parent)
        self.buttons = QtWidgets.QDialogButtonBox(parent)

        self.copy.setText('To Clipboard')
        self.save.setText('To File')
        self.show_details.setText('Details...')

        self.layout.addWidget(self.icon, 0, 0, 2, 1)
        self.layout.addWidget(self.message, 0, 1, 1, 4)
        self.layout.addWidget(self.details, 1, 1, 1, 4)
        self.layout.addWidget(self.copy, 2, 1)
        self.layout.addWidget(self.save, 2, 2)
        self.layout.addWidget(self.show_details, 2, 3)
        self.layout.addWidget(self.buttons, 2, 4)

        self.layout.setColumnStretch(4, 1)


class Dialog(QtWidgets.QDialog):
    def __init__(self, *args, cancellable=False, details=False,
                 save_filters=None, save_caption=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.save_filters = save_filters
        if self.save_filters is None:
            self.save_filters = (
                ('Text', ['txt']),
                ('All Files', ['*'])
            )

        self.save_caption = save_caption

        self.ui = DialogUi(parent=self)

        self.ui.details.setVisible(False)
        self.ui.show_details.setVisible(details)

        self.ui.buttons.accepted.connect(self.accept)
        self.ui.buttons.rejected.connect(self.reject)

        self.ui.copy.clicked.connect(self.copy)
        self.ui.save.clicked.connect(self.save)
        self.ui.show_details.clicked.connect(self.show_details)

        self.setLayout(self.ui.layout)
        buttons = QtWidgets.QDialogButtonBox.Ok
        if cancellable:
            buttons |= QtWidgets.QDialogButtonBox.Cancel

        self.ui.buttons.setStandardButtons(buttons)

        self.text = None
        self.html = None
        self.details_text = None
        self.details_html = None

        desktops = QtWidgets.QApplication.desktop()
        screen_number = desktops.screenNumber(self.parent())
        geometry = desktops.screenGeometry(screen_number)

        self.setMaximumHeight(geometry.height() * 0.7)
        self.setMaximumWidth(geometry.width() * 0.7)
        self.minimum_size = self.minimumSize()
        self.maximum_size = self.maximumSize()

    def all_as_text(self):
        message = self.ui.message.toPlainText().strip()
        details = self.ui.details.toPlainText().strip()

        if len(details) == 0:
            return message + '\n'

        return textwrap.dedent('''\
            {message}

             - - - - Details:

            {details}
            '''
                            ).format(
            message=message,
            details=details,
        )

    def copy(self):
        QtWidgets.QApplication.clipboard().setText(self.all_as_text())

    def save(self):
        extras = {}
        if self.save_caption is not None:
            extras['caption'] = self.save_caption

        path = epyqlib.utils.qt.file_dialog(
            filters=self.save_filters,
            parent=self,
            save=True,
            **extras,
        )

        if path is None:
            return

        with open(path, 'w') as f:
            f.write(self.all_as_text())

    def show_details(self):
        to_be_visible = not self.ui.details.isVisible()
        self.ui.details.setVisible(to_be_visible)
        self.set_size()

    def set_size(self):
        self.setFixedSize(self.sizeHint())
        self.setMinimumSize(self.minimum_size)
        self.setMaximumSize(self.maximum_size)

    def set_text(self, text):
        self.ui.message.setPlainText(text)

        self.html = None
        self.text = text

        self.ui.message.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)

    def set_html(self, html):
        self.ui.message.setHtml(html)

        self.html = html
        self.text = None

        self.ui.message.setLineWrapMode(QtWidgets.QTextEdit.WidgetWidth)

    def set_details_text(self, text):
        self.ui.details.setPlainText(text)

        self.details_html = None
        self.details_text = text

        self.ui.details.setLineWrapMode(QtWidgets.QTextEdit.NoWrap)

    def set_details_html(self, html):
        self.ui.details.setHtml(html)

        self.details_html = html
        self.details_text = None

        self.ui.details.setLineWrapMode(QtWidgets.QTextEdit.WidgetWidth)

    def set_message_box_icon(self, icon):
        self.ui.icon.setPixmap(QtWidgets.QMessageBox.standardIcon(icon))


def dialog(parent, message, title=None, icon=None,
           rich_text=False, details='', details_rich_text=False,
           cancellable=False, modal=True, **kwargs):
    box = Dialog(
        parent=parent,
        cancellable=cancellable,
        details=len(details) > 0,
        **kwargs,
    )

    box.setModal(modal)

    if rich_text:
        box.set_html(message)
    else:
        box.set_text(message)

    if details_rich_text:
        box.set_details_html(details)
    else:
        box.set_details_text(details)

    if icon is not None:
        box.set_message_box_icon(icon)

    if title is not None:
        parent_title = QtWidgets.QApplication.instance().applicationName()

        if len(parent_title) > 0:
            title = ' - '.join((
                parent_title,
                title,
            ))


        box.setWindowTitle(title)

    box.finished.connect(box.deleteLater)

    if modal:
        return box.exec()

    box.show()
    return


def dialog_from_file(parent, title, file_name):
    # The Qt Installer Framework (QtIFW) likes to do a few things to license files...
    #  * '\n' -> '\r\n'
    #   * even such that '\r\n' -> '\r\r\n'
    #  * Recodes to something else (probably cp-1251)
    #
    # So, we'll just try different encodings and hope one of them works.

    encodings = [None, 'utf-8']

    for encoding in encodings:
        try:
            with open(os.path.join('Licenses', file_name), encoding=encoding) as in_file:
                message = in_file.read()
        except UnicodeDecodeError:
            pass
        else:
            break

    dialog(
        parent=parent,
        title=title,
        message=message,
    )


class PySortFilterProxyModel(QtCore.QSortFilterProxyModel):
    def __init__(self, *args, filter_column, **kwargs):
        super().__init__(*args, **kwargs)

        # TODO: replace with filterKeyColumn
        self.filter_column = filter_column

        self.wildcard = QtCore.QRegExp()
        self.wildcard.setPatternSyntax(QtCore.QRegExp.Wildcard)

    def lessThan(self, left, right):
        left_model = left.model()
        left_data = (
            left_model.data(left, self.sortRole())
            if left_model else
            None
        )

        right_model = right.model()
        right_data = (
            right_model.data(right, self.sortRole())
            if right_model else
            None
        )

        return left_data < right_data

    def filterAcceptsRow(self, row, parent):
        # TODO: do i need to invalidate any time i set the 'regexp'?
        # http://doc.qt.io/qt-5/qsortfilterproxymodel.html#invalidateFilter

        pattern = self.filterRegExp().pattern()
        if pattern == '':
            return True

        pattern = '*{}*'.format(pattern)

        model = self.sourceModel()
        result = False
        index = model.index(row, self.filter_column, parent)
        self_index = self.index(row, self.filter_column, parent)
        result |= self.hasChildren(self_index)
        self.wildcard.setPattern(pattern)
        result |= self.wildcard.exactMatch(model.data(index, QtCore.Qt.DisplayRole))

        return result

    def next_row(self, index):
        return self.sibling(
            index.row() + 1,
            index.column(),
            index,
        )

    def next_index(self, index, allow_children=True):
        if allow_children and self.hasChildren(index):
            return self.index(0, index.column(), index), False

        next_ = self.next_row(index)
        if not next_.isValid():
            while True:
                index = index.parent()
                if not index.isValid():
                    return self.index(0, 0, QtCore.QModelIndex()), True

                next_ = self.next_row(index)
                if next_.isValid():
                    break

        return next_, False

    def search(self, text, search_from, column):
        def set_row_column(index, row=None, column=None):
            if row is None:
                row = index.row()

            if column is None:
                column = index.column()

            return self.index(
                row,
                column,
                index.parent(),
            )

        if text == '':
            return None

        text = '*{}*'.format(text)

        flags = (
            QtCore.Qt.MatchContains
            | QtCore.Qt.MatchRecursive
            | QtCore.Qt.MatchWildcard
        )

        wrapped = False

        if search_from.isValid():
            search_from, wrapped = self.next_index(search_from)
        else:
            search_from = self.index(0, 0, QtCore.QModelIndex())

        while True:
            next_indexes = self.match(
                set_row_column(index=search_from, column=column),
                QtCore.Qt.DisplayRole,
                text,
                1,
                flags,
            )

            if len(next_indexes) > 0:
                next_index, = next_indexes

                if not next_index.isValid():
                    break

                return next_index
            elif wrapped:
                break

            search_from, wrapped = self.next_index(search_from)

        # TODO: report not found and/or wrap
        print('reached end')
        return None


@attr.s
class DiffProxyModel(QtCore.QIdentityProxyModel):
    parent = attr.ib(default=None)
    columns = attr.ib(factory=set, converter=set)
    _reference_column = attr.ib(default=None)
    diff_highlights = attr.ib(factory=dict)
    reference_highlights = attr.ib(factory=dict)
    diff_role = attr.ib(default=QtCore.Qt.ItemDataRole.DisplayRole)

    def __attrs_post_init__(self):
        super().__init__(self.parent)

    def data(self, index, role):
        column = index.column()

        if self.reference_column is not None:
            if (
                column == self.reference_column
                and role in self.reference_highlights
            ):
                return self.reference_highlights[role]
            elif (
                column != self.reference_column
                and role in self.diff_highlights
                and column in self.columns
            ):
                this_value = super().data(
                    index,
                    self.diff_role,
                )
                that_value = super().data(
                    index.siblingAtColumn(self.reference_column),
                    self.diff_role,
                )

                if this_value != that_value:
                    return self.diff_highlights[role]

        return super().data(index, role)

    @property
    def reference_column(self):
        return self._reference_column

    @reference_column.setter
    def reference_column(self, column):
        self._reference_column = column
        self.all_changed()

    def column_group_limits(self):
        return [
            (group[0], group[-1])
            for group in epyqlib.utils.general.contiguous_groups(
                sorted(self.columns),
            )
        ]

    def roles(self):
        return {*self.diff_highlights, *self.reference_highlights}

    def all_changed(self):
        indexes = [QtCore.QModelIndex()]

        while len(indexes) > 0:
            parent = indexes.pop()
            if self.hasChildren(parent):
                row_count = self.rowCount(parent)

                indexes.extend([
                    self.index(row, 0, parent)
                    for row in range(row_count)
                ])

                for start, end in self.column_group_limits():
                    self.dataChanged.emit(
                        self.index(0, start, parent),
                        self.index(row_count - 1, end, parent),
                        self.roles(),
                    )

    def setData(self, index, value, role):
        changed = super().setData(index, value)

        if changed and role == self.diff_role:
            column = index.column()
            if column == self.reference_column:
                for start, end in self.column_group_limits():
                    self.dataChanged.emit(
                        index.siblingAtColumn(start),
                        index.siblingAtColumn(end),
                        self.diff_highlights,
                    )
            elif column in self.columns:
                self.dataChanged.emit(index, index, self.diff_highlights)

        return changed


def load_ui(filepath, base_instance):
    # TODO: CAMPid 9549757292917394095482739548437597676742
    ui_file = QtCore.QFile(filepath)
    ui_file.open(QtCore.QFile.ReadOnly | QtCore.QFile.Text)
    ts = QtCore.QTextStream(ui_file)
    sio = io.StringIO(ts.readAll())

    return PyQt5.uic.loadUi(sio, base_instance)


def search_view(view, text, column):
    if text == '':
        return

    model = view.model()

    models = []

    while model is not None:
        models.append(model)
        search = getattr(model, 'search', None)
        if search is not None:
            break

        model = model.sourceModel()
    else:
        raise Exception('ack')

    index = search(
        text=text,
        column=column,
        search_from=view.currentIndex(),
    )

    if index is not None:
        parent = index.parent()
        # TODO: not sure why but this must be set to zero or the row
        #       won't be highlighted.  it still gets expanded and printing
        #       the display role data still works.
        parent = model.index(parent.row(), 0, parent.parent())
        index = model.index(index.row(), index.column(), parent)

        for model in reversed(models[:-1]):
            index = model.mapFromSource(index)

        view.setCurrentIndex(index)
        view.selectionModel().select(
            index,
            QtCore.QItemSelectionModel.ClearAndSelect,
        )


class NotAPyQtifyInstance(Exception):
    pass


def pyqtified(instance):
    return instance.__pyqtify_instance__


@attr.s
class PyQtifyInstance:
    display_name = attr.ib()
    values = attr.ib(default={})
    changed = attr.ib(default=None)

    @classmethod
    def fill(cls, display_name, attrs_class):
        return cls(
            display_name=display_name,
            values={
                field.name: None
                for field in attr.fields(attrs_class)
            },
        )


def pyqtify(name=None, property_decorator=lambda: property):
    def inner(cls):
        if name is None:
            display_name = cls.__name__
        else:
            display_name = name

        names = tuple(field.name for field in attr.fields(cls))

        def __getitem__(self, key):
            if key not in self.names:
                raise KeyError(key)

            return getattr(self, signal_name(key))

        def __getattr__(self, name):
            if name not in self.names:
                raise AttributeError(
                    "'{class_name}' object has no attribute '{name}'".format(
                        class_name=type(self).__name__,
                        attribute=name,
                    )
                )

            return getattr(self, signal_name(name))

        def signal_name(name):
            return '_pyqtify_signal_{}'.format(name)

        SignalContainer = type(
            'SignalContainer',
            (PyQt5.QtCore.QObject,),
            {
                'names': names,
                '__getattr__': __getattr__,
                '__getitem__': __getitem__,
                **{
                    signal_name(name): PyQt5.QtCore.pyqtSignal('PyQt_PyObject')
                    for name in names
                },
            },
        )

        old_init = cls.__init__

        def __init__(self, *args, **kwargs):
            self.__pyqtify_instance__ = PyQtifyInstance.fill(
                display_name=display_name,
                attrs_class=type(self),
            )

            self.__pyqtify_instance__.changed = SignalContainer()

            try:
                old_init(self, *args, **kwargs)
            except TypeError as e:
                raise TypeError(
                    '.'.join((
                        type(self).__module__,
                        type(self).__qualname__,
                        e.args[0],
                    )),
                ) from e

        cls.__init__ = __init__

        for name_ in names:
            property_ = getattr(cls, 'pyqtify_{}'.format(name_), None)

            if property_ is None:
                @property_decorator()
                def property_(self, name=name_):
                    return pyqtify_get(self, name)

                @property_.setter
                def property_(self, value, name=name_):
                    pyqtify_set(self, name, value)

            setattr(cls, name_, property_)

        return cls

    return inner


def pyqtify_get(instance, name):
    return instance.__pyqtify_instance__.values[name]


def pyqtify_set(instance, name, value):
    if value != instance.__pyqtify_instance__.values[name]:
        instance.__pyqtify_instance__.values[name] = value
        try:
            instance.__pyqtify_instance__.changed[name].emit(value)
        except RuntimeError:
            pass


def pyqtify_signals(instance):
    try:
        instance = instance.__pyqtify_instance__
    except AttributeError as e:
        raise NotAPyQtifyInstance from e

    return instance.changed


def pyqtify_passthrough_properties(original, field_names):
    def inner(cls):
        old_init = cls.__init__

        def __init__(self, *args, **kwargs):
            old_init(self, *args, **kwargs)

            original_object = getattr(self, original)
            signals = epyqlib.utils.qt.pyqtify_signals(self)

            def original_changed(new_original, self=self):
                # TODO: need to be disconnecting as well
                signals = epyqlib.utils.qt.pyqtify_signals(self)

                try:
                    new_original_signals = (
                        epyqlib.utils.qt.pyqtify_signals(new_original)
                    )
                except NotAPyQtifyInstance:
                    pass
                else:
                    for name in field_names:
                        new_original_signals[name].connect(signals[name])

            getattr(signals, original).connect(original_changed)
            original_changed(original_object)

        cls.__init__ = __init__

        for name in field_names:
            @property
            def property_(self, name=name):
                original_ = getattr(self, original)
                if original_ is None or isinstance(original_, uuid.UUID):
                    return pyqtify_get(self, name)

                return getattr(original_, name)

            @property_.setter
            def property_(self, value, name=name):
                original_ = getattr(self, original)
                if original_ is None:
                    pyqtify_set(self, name, value)
                else:
                    setattr(original_, name, value)

            setattr(cls, 'pyqtify_' + name, property_)

        return cls

    return inner


class TargetModelNotReached(Exception):
    pass


def resolve_models(model, target=None):
    sentinel = object()
    if target is None:
        target = sentinel

    models = [model]

    while isinstance(models[-1], PyQt5.QtCore.QAbstractProxyModel):
        models.append(models[-1].sourceModel())

        if models[-1] is target:
            break

    if target is not sentinel and models[-1] is not target:
        raise TargetModelNotReached()

    return models


def resolve_index_to_model(index, target=None):
    model = index.model()

    if model is target:
        return index

    model_pairs = epyqlib.utils.general.pairwise(
        resolve_models(model=model, target=target),
    )

    for first, second in model_pairs:
        index = first.mapToSource(index)

        if second is target:
            return index

    return index


def resolve_index_from_model(model, view, index):
    models = resolve_models(model=view.model(), target=model)

    for model in models[-2::-1]:
        index = model.mapFromSource(index)

    return index


def sigint_handler(signal_number, stack_frame):
    QtWidgets.QApplication.exit(128 + signal_number)


def setup_sigint():
    signal.signal(signal.SIGINT, sigint_handler)

    # Regularly give Python a chance to receive signals such as ctrl+c
    timer = QtCore.QTimer()
    timer.start(200)
    timer.timeout.connect(lambda: None)

    return timer


# class SignalWrapper(PyQt5.QtCore.pyqtBoundSignal):
#     def __init__(self, owner, wrapped):
#         self.wrapped = wrapped
#         self.owner = owner
#
#     def __getattribute__(self, item):
#         if item == '__repr__':
#             return super().__getattribute__(item)
#
#         wrapped = super().__getattribute__('wrapped')
#         return getattr(wrapped, item)
#
#     def __repr__(self):
#         owner = super().__getattribute__('owner')
#         wrapped = super().__getattribute__('wrapped')
#
#         return '<{} of {}>'.format(
#             repr(wrapped)[1:-1],
#             repr(owner).rsplit('.', 1)[1][:-1],
#         )
#
#
# def signal_repr(self):
#     owner = super().__getattribute__('owner')
#     wrapped = super().__getattribute__('wrapped')
#
#     return '<{} of {}>'.format(
#         repr(wrapped)[1:-1],
#         repr(owner).rsplit('.', 1)[1][:-1],
#     )


class Signal:
    attribute_name = None

    def __init__(self, *args, **kwargs):
        class _SignalQObject(QtCore.QObject):
            signal = QtCore.pyqtSignal(*args, **kwargs)

        self.object_cls = _SignalQObject

    def __get__(self, instance, owner):
        if instance is None:
            return self

        d = getattr(instance, self.attribute_name, None)

        if d is None:
            d = {}
            setattr(instance, self.attribute_name, d)

        o = d.get(self.object_cls)
        if o is None:
            o = self.object_cls()
            d[self.object_cls] = o

        signal = o.signal
        # signal.__repr__ = signal_repr
        # return SignalWrapper(owner=instance, wrapped=o.signal)
        return signal

    def qobject_host(self, instance):
        """Return the QObject which hosts the pyqtSignal on the passed instance.
                
        ``TheClass.the_signal.qobject_host(an_instance)`` will return the ``QObject``
        instance used to host the signal ``an_instance.the_signal``.
        """
        return getattr(instance, self.attribute_name)[self.object_cls]


Signal.attribute_name = epyqlib.utils.general.identifier_path(Signal)


class Connections:
    def __init__(self, signal, slot=None, slots=(), connect=True):
        self.signal = signal
        self.slots = slots

        if slot is not None:
            self.slots = (slot,) + self.slots

        if connect:
            self.connect()

    def connect(self):
        for slot in self.slots:
            self.signal.connect(slot)

    def disconnect(self):
        for slot in self.slots:
            self.signal.disconnect(slot)


def set_expanded_tree(view, index, expanded):
    if not expanded:
        view.setExpanded(index, expanded)

    for row in range(index.model().rowCount(index)):
        set_expanded_tree(
            view=view,
            index=index.child(row, 0),
            expanded=expanded,
        )

    if expanded:
        view.setExpanded(index, expanded)


@enum.unique
class UserRoles(epyqlib.utils.general.AutoNumberIntEnum):
    unique = QtCore.Qt.UserRole
    sort = None
    raw = None
    node = None
    field_name = None
    attrs_model = None
    column_index = None


def child_text_list_from_index(index, model, recurse=True):
    row_count = model.rowCount(index)
    if row_count == 0:
        return []

    lines = []
    for i in range(row_count):
        child = model.index(i, 0, index)
        lines.append(child.data())
        if recurse:
            lines.append(child_text_list_from_index(index=child, model=model))

    return lines


def indented_text_from_model(model, index=None):
    if index is None:
        index = QtCore.QModelIndex()

    lines = child_text_list_from_index(
        index=index,
        model=model,
        recurse=True,
    )
    lines = [
        'Root',
        lines,
    ]

    return epyqlib.utils.general.format_nested_lists(lines, indent="'   ")


@attr.s
class DeferredForSignal:
    signal = attr.ib()
    deferred = attr.ib(
        default=attr.Factory(
            factory=lambda self: twisted.internet.defer.Deferred(
                canceller=self.cancelled,
            ),
            takes_self=True,
        ),
    )
    timeout_call = attr.ib(default=None)

    def connect(self, timeout=None):
        self.signal.connect(self.slot)

        if timeout is not None:
            import twisted.internet.reactor
            self.timeout_call = twisted.internet.reactor.callLater(
                timeout,
                self.time_out,
            )

    def time_out(self):
        self.signal.disconnect(self.slot)
        self.deferred.errback(epyqlib.utils.twisted.RequestTimeoutError())

    def disconnect(self):
        self.signal.disconnect(self.slot)
        if self.timeout_call is not None:
            self.timeout_call.cancel()

    def cancelled(self, deferred):
        self.disconnect()

    def slot(self, *args):
        self.disconnect()
        self.deferred.callback(args)


def signal_as_deferred(signal, timeout=None, f=None, *args, **kwargs):
    dfs = DeferredForSignal(signal=signal)
    dfs.connect(timeout=timeout)

    if f is not None:
        f(*args, **kwargs)

    return dfs.deferred



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'device.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(824, 809)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setObjectName("verticalLayout")
        self.overlay_stack = QtWidgets.QStackedWidget(Form)
        self.overlay_stack.setObjectName("overlay_stack")
        self.overlays = QtWidgets.QWidget()
        self.overlays.setObjectName("overlays")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.overlays)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.offline_overlay = OverlayLabel(self.overlays)
        self.offline_overlay.setObjectName("offline_overlay")
        self.gridLayout_3.addWidget(self.offline_overlay, 0, 0, 1, 1)
        self.connection_monitor_overlay = OverlayLabel(self.overlays)
        self.connection_monitor_overlay.setObjectName("connection_monitor_overlay")
        self.gridLayout_3.addWidget(self.connection_monitor_overlay, 1, 0, 1, 1)
        self.overlay_stack.addWidget(self.overlays)
        self.main_stack = QtWidgets.QWidget()
        self.main_stack.setObjectName("main_stack")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.main_stack)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.tabs = QtWidgets.QTabWidget(self.main_stack)
        self.tabs.setObjectName("tabs")
        self.txrx = QtWidgets.QWidget()
        self.txrx.setObjectName("txrx")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.txrx)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter = QtWidgets.QSplitter(self.txrx)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        self.splitter.setObjectName("splitter")
        self.layoutWidget = QtWidgets.QWidget(self.splitter)
        self.layoutWidget.setObjectName("layoutWidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.layoutWidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(self.layoutWidget)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.tx = TxRxView(self.layoutWidget)
        self.tx.setObjectName("tx")
        self.verticalLayout_2.addWidget(self.tx)
        self.layoutWidget_2 = QtWidgets.QWidget(self.splitter)
        self.layoutWidget_2.setObjectName("layoutWidget_2")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.layoutWidget_2)
        self.verticalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.label_2 = QtWidgets.QLabel(self.layoutWidget_2)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.verticalLayout_5.addWidget(self.label_2)
        self.rx = TxRxView(self.layoutWidget_2)
        self.rx.setObjectName("rx")
        self.verticalLayout_5.addWidget(self.rx)
        self.verticalLayout_3.addWidget(self.splitter)
        self.tabs.addTab(self.txrx, "")
        self.variables = QtWidgets.QWidget()
        self.variables.setObjectName("variables")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.variables)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.variable_selection = VariableSelection(self.variables)
        self.variable_selection.setObjectName("variable_selection")
        self.verticalLayout_6.addWidget(self.variable_selection)
        self.tabs.addTab(self.variables, "")
        self.nv = QtWidgets.QWidget()
        self.nv.setObjectName("nv")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.nv)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.nv1 = NvView(self.nv)
        self.nv1.setObjectName("nv1")
        self.verticalLayout_4.addWidget(self.nv1)
        self.tabs.addTab(self.nv, "")
        self.scripting = QtWidgets.QWidget()
        self.scripting.setObjectName("scripting")
        self.gridLayout = QtWidgets.QGridLayout(self.scripting)
        self.gridLayout.setObjectName("gridLayout")
        self.scripting_view = ScriptingView(self.scripting)
        self.scripting_view.setObjectName("scripting_view")
        self.gridLayout.addWidget(self.scripting_view, 0, 0, 1, 1)
        self.tabs.addTab(self.scripting, "")
        self.faultlog = QtWidgets.QWidget()
        self.faultlog.setObjectName("faultlog")
        self.gridLayout1 = QtWidgets.QGridLayout(self.faultlog)
        self.gridLayout1.setObjectName("gridLayout1")
        self.fault_log_view = FaultLogView(self.faultlog)
        self.fault_log_view.setObjectName("fault_log_view")
        self.gridLayout1.addWidget(self.fault_log_view, 0, 0, 1, 1)
        self.tabs.addTab(self.faultlog, "")
        self.files = QtWidgets.QWidget()
        self.files.setObjectName("files")
        self.gridLayout2 = QtWidgets.QGridLayout(self.files)
        self.gridLayout2.setObjectName("gridLayout2")
        self.files_view = FilesViewQtBuilder(self.files)
        self.files_view.setObjectName("files_view")
        self.gridLayout2.addWidget(self.files_view, 0, 0, 1, 1)
        self.tabs.addTab(self.files, "")
        self.gridLayout_2.addWidget(self.tabs, 0, 0, 1, 1)
        self.overlay_stack.addWidget(self.main_stack)
        self.verticalLayout.addWidget(self.overlay_stack)

        self.retranslateUi(Form)
        self.tabs.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.offline_overlay.setProperty("text", _translate("Form", "offline"))
        self.label.setText(_translate("Form", "Transmit"))
        self.label_2.setText(_translate("Form", "Receive"))
        self.tabs.setTabText(self.tabs.indexOf(self.txrx), _translate("Form", "Tx/Rx"))
        self.tabs.setTabText(self.tabs.indexOf(self.variables), _translate("Form", "Data Logger"))
        self.tabs.setTabText(self.tabs.indexOf(self.nv), _translate("Form", "Parameters"))
        self.tabs.setTabText(self.tabs.indexOf(self.scripting), _translate("Form", "Scripting"))
        self.tabs.setTabText(self.tabs.indexOf(self.faultlog), _translate("Form", "Fault Log"))
        self.tabs.setTabText(self.tabs.indexOf(self.files), _translate("Form", "Files"))
from epyqlib.faultlogview import FaultLogView
from epyqlib.filesview import FilesViewQtBuilder
from epyqlib.nvview import NvView
from epyqlib.overlaylabel import OverlayLabel
from epyqlib.scriptingview import ScriptingView
from epyqlib.txrxview import TxRxView
from epyqlib.variableselection import VariableSelection



from PyQt5 import QtWidgets
from PyQt5.QtCore import QSortFilterProxyModel

import epyqlib.txrxview_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = 'Copyright 2016, EPC Power Corp.'
__license__ = 'GPLv2+'


class TxRxView(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.txrxview_ui.Ui_Form()
        self.ui.setupUi(self)

        self.resize_columns = epyqlib.txrx.Columns.fill(False)

    def setModel(self, model):
        self.ui.tree_view.setModel(model)

        tx = self.nonproxy_model().root.tx
        if tx:
            self.ui.searchbox.connect_to_view(
                view=self.ui.tree_view,
                column=epyqlib.txrx.Columns.indexes.name,
            )

        self.ui.searchbox.setHidden(not tx)

        self.ui.tree_view.header().setStretchLastSection(False)

        for i in epyqlib.txrx.Columns.indexes:
            if self.resize_columns[i]:
                self.ui.tree_view.header().setSectionResizeMode(
                    i, QtWidgets.QHeaderView.ResizeToContents)
            else:
                # at least fit the column headers and/or initial data
                self.ui.tree_view.resizeColumnToContents(i)

        self.ui.tree_view.header().setSectionResizeMode(
            epyqlib.txrx.Columns.indexes.name, QtWidgets.QHeaderView.Stretch)

        self.ui.tree_view.setItemDelegateForColumn(
            epyqlib.txrx.Columns.indexes.value,
            epyqlib.delegates.ByFunction(
                model=self.nonproxy_model(),
                parent=self,
            )
        )

        self.ui.tree_view.setColumnWidth(epyqlib.txrx.Columns.indexes.value,
                                         self.calculate_max_value_width())
        self.ui.tree_view.setColumnWidth(epyqlib.txrx.Columns.indexes.id,
                                         self.calculate_max_id_width() +
                                         self.ui.tree_view.indentation())

    # TODO: CAMPid 07943342700734207878034207087
    def nonproxy_model(self):
        model = self.ui.tree_view.model()
        while isinstance(model, QSortFilterProxyModel):
            model = model.sourceModel()

        return model

    # TODO: CAMPid 989849193479134917954791341
    def calculate_max_value_width(self):
        metric = self.ui.tree_view.fontMetrics()
        chars = ['{:X}'.format(i) for i in range(16)]
        widths = [metric.width(c) for c in chars]
        widest_width = max(widths)
        widest_char = chars[widths.index(widest_width)]
        string = ' '.join([widest_char * 2] * 8)
        return metric.width(string)

    # TODO: CAMPid 989849193479134917954791341
    def calculate_max_id_width(self):
        metric = self.ui.tree_view.fontMetrics()
        chars = ['{:X}'.format(i) for i in range(16)]
        widths = [metric.width(c) for c in chars]
        widest_width = max(widths)
        widest_char = chars[widths.index(widest_width)]
        string = '0x{}'.format(widest_char * 8)
        return metric.width(string)

    def set_sorting_enabled(self, enabled):
        self.ui.tree_view.setSortingEnabled(enabled)

    def sort_by_column(self, column, order):
        self.ui.tree_view.sortByColumn(column, order)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'iogroup.ui'
#
# Created by: PyQt5 UI code generator 5.13.0
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(909, 590)
        self.verticalLayout = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.box = QtWidgets.QGroupBox(Form)
        self.box.setObjectName("box")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.box)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setObjectName("layout")
        self.verticalLayout_3.addLayout(self.layout)
        self.verticalLayout.addWidget(self.box)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.box.setTitle(_translate("Form", "GroupBox"))
